<xsl:stylesheet version="2.0"
                xmlns:xsl="http://www.w3.org/1999/XSL/Transform"
                xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance"
                xmlns:clearsy="http://clearsy">
  <xsl:output method="xml"/>

  <!-- Note:
       Certaines constructions xsl ne doivent pas etre utilisées dans ce document
       car elles sont incorrectement implémentées dans Qt:
       (cf https://gitlab.clearsy.com/AtelierB/specification-pog/issues/12
       - la balise <xsl:variable> à l'intérieur d'une balise <xsl:for-each>
       - la balise <xsl:apply-template> avec des parametres (<xsl:with-param>).
  -->

  <!-- Factorisation of the definition of & operator
  -->
  <xsl:variable name="and"><![CDATA[&]]></xsl:variable>

  <!-- Define some common hypothesis in define elements.
       Those defines are set in a template in order to call them each time it is necessary.
  -->
  <xsl:template name="common_part">
    <!-- Define "B definitions": element containing the two definitions for NAT and INT.
         NAT = 0..MAXINT &
         INT = MININT..MAXINT
    -->
    <Define name="B definitions">
      <Nary_Pred>
        <xsl:attribute name="op">
          <xsl:value-of select="$and" />
        </xsl:attribute>
        <Exp_Comparison op="=">
          <Id value="NAT" typref="2"/>
          <Binary_Exp op=".." typref="2">
            <Integer_Literal value="0" typref="1"/>
            <Id value="MAXINT" typref="1" />
          </Binary_Exp>
        </Exp_Comparison>
        <Exp_Comparison op="=">
	  <Id value="INT" typref="2"/>
	  <Binary_Exp op=".." typref="2">
            <Id value="MININT" typref="1"/>
            <Id value="MAXINT" typref="1"/>
          </Binary_Exp>
        </Exp_Comparison>
      </Nary_Pred>
    </Define>

    <!-- Define "ctx": element containing Sets and Properties from Seen
         component and their included component
    -->
    <Define name="ctx">
      <xsl:copy-of select="Sees/Machine/Sets/*"/>
      <xsl:copy-of select="Sees/Machine/Includes/Machine/Sets/*"/>
      <xsl:copy-of select="Sees/Machine/Properties/*"/>
      <xsl:copy-of select="Sees/Machine/Includes/Machine/Properties/*"/>
    </Define>

    <!-- Define "seext": element containing Invariant and Assertions from Seen
         component and their included component
    -->
    <Define name="seext">
      <xsl:copy-of select="Sees/Machine/Invariant/*"/>
      <xsl:copy-of select="Sees/Machine/Assertions/*"/>
      <xsl:copy-of select="Sees/Machine/Includes/Machine/Invariant/*"/>
      <xsl:copy-of select="Sees/Machine/Includes/Machine/Assertions/*"/>
    </Define>
  </xsl:template>

  <!-- Define some common hypothesis in define elements.
       Those defines are set in a template in order to call them each time it is necessary.
  -->
  <xsl:template name="common_part2">
    <!-- Define "inv": element containing Invariant of the current component
    -->
    <Define name="inv">
      <xsl:copy-of select="Invariant/*"/>
    </Define>

    <!-- Define "ass": element containing Assertions of the current component
    -->
    <Define name="ass">
      <xsl:copy-of select="Assertions/*"/>
    </Define>
  </xsl:template>

  <!-- Define some common hypothesis in define elements used in machine proof obligations.
       Those defines are set in a template in order to call them each time it is necessary.
  -->
  <xsl:template name="common_part_of_specification">
    <!-- Define "lprp": element containing Sets and Properties of the current machine
    -->
    <Define name="lprp">
      <xsl:copy-of select="Sets/*"/>
      <xsl:copy-of select="Properties/*"/>
    </Define>

    <!-- Define "inprp": element containing Sets and Properties of included components of the current machine
    -->
    <Define name="inprp">
      <xsl:copy-of select="Includes/Machine/Sets/*"/>
      <xsl:copy-of select="Includes//Machine/Properties/*"/>
    </Define>

    <!-- Define "inext": element containing Invariant and Assertions of included components of the current machine
    -->
    <Define name="inext">
      <xsl:copy-of select="Includes/Machine/Invariant/*"/>
      <xsl:copy-of select="Includes/Machine/Assertions/*"/>
    </Define>
  </xsl:template>

  <!-- Define some common hypothesis in define elements used in refinement proof obligations.
       Those defines are set in a template in order to call them each time it is necessary.
  -->
  <xsl:template name="common_part_of_refinement">
    <!-- Define "mchcst": element containing the Constraint clause of the machine refined by the current refinement
         last()
    -->
    <Define name="mchcst">
      <xsl:copy-of select="Abstraction[last()]/Machine/Constraints/*"/>
    </Define>

    <Define name="aprp">
      <xsl:copy-of select="Abstraction/Machine/Sets/*"/>
      <xsl:copy-of select="Abstraction/Machine/Includes/Machine/Sets/*"/>
      <xsl:copy-of select="Abstraction/Machine/Properties/*"/>
      <xsl:copy-of select="Abstraction/Machine/Includes/Machine/Properties/*"/>
    </Define>

    <Define name="abs">
      <xsl:copy-of select="Abstraction/Machine/Invariant/*"/>
      <xsl:copy-of select="Abstraction/Machine/Assertions/*"/>
      <xsl:copy-of select="Abstraction/Machine/Includes/Machine/Invariant/*"/>
      <xsl:copy-of select="Abstraction/Machine/Includes/Machine/Assertions/*"/>
    </Define>
  </xsl:template>

  <xsl:template match="Machine[@type='abstraction']">
    <Proof_Obligations>

      <xsl:call-template name="common_part"/>
      <xsl:call-template name="common_part_of_specification"/>
      <xsl:call-template name="common_part2"/>

      <Define name="cst">
        <xsl:copy-of select="Constraints/*"/>
      </Define>

      <Define name="sets">
	<xsl:copy-of select="Sets/*"/>
      </Define>

      <xsl:for-each select="Includes/*">
        <Proof_Obligation>
          <Comment>Inclusion dans une machine abstraite</Comment>
          <Tag>InstanciatedConstraintsLemmas</Tag>
          <Definition name="B definitions"/>
          <Definition name="ctx"/>
          <Definition name="cst"/>
          <Definition name="lprp"/>
	  <Goal goalTag="Instanciation of included machine">
            <Nary_Pred>
              <xsl:attribute name="op">
                <xsl:value-of select="$and" />
              </xsl:attribute>
              <xsl:copy-of select="Constraints/*"/>
            </Nary_Pred>
          </Goal>
        </Proof_Obligation>
      </xsl:for-each>

      <xsl:for-each select="Assertions/*">
        <xsl:call-template name="machine_Assertions">
          <xsl:with-param name="pos" select="position()"/>
        </xsl:call-template>
      </xsl:for-each>

      <Proof_Obligation>
        <Comment>Initialisation dans une machine abstraite</Comment>
        <Tag>Initialisation</Tag>
        <Definition name="B definitions"/>
        <Definition name="ctx"/>
        <Definition name="cst"/>
        <Definition name="lprp"/>
        <Definition name="inprp"/>
        <Definition name="inext"/>
        <Definition name="seext"/>
        <Goal>
	  <xsl:attribute name="goalTag">Invariant is preserved</xsl:attribute>
          <Sub_Calculus>
            <Nary_Sub op=";">
              <xsl:for-each select="Includes">
                <xsl:copy-of select="Machine/Initialisation/*"/>
              </xsl:for-each>
              <xsl:for-each select="Initialisation">
                <xsl:copy-of select="./*"/>
              </xsl:for-each>
            </Nary_Sub>
            <Binary_Pred op="=>">
              <Nary_Pred>
                <xsl:attribute name="op">
                  <xsl:value-of select="$and" />
                </xsl:attribute>
                <xsl:copy-of select="Includes/Machine/Invariant/*"/>
                <xsl:copy-of select="Includes/Machine/Assertions/*"/>
	      </Nary_Pred>
	      <Nary_Pred>
		<xsl:attribute name="op">
		  <xsl:value-of select="$and" />
		</xsl:attribute>
		<xsl:copy-of select="Invariant/*"/>
	      </Nary_Pred>
	    </Binary_Pred>
          </Sub_Calculus>
        </Goal>
      </Proof_Obligation>

      <xsl:for-each select="Operations/*">
        <Proof_Obligation>
          <Comment>Operations dans une machine abstraite</Comment>
          <Tag><xsl:value-of select="@name"/></Tag>
          <Definition name="B definitions"/>
          <Definition name="ctx"/>
          <Definition name="cst"/>
          <Definition name="lprp"/>
          <Definition name="inprp"/>
          <Definition name="inext"/>
          <Definition name="seext"/>
          <Definition name="inv"/>
          <Definition name="ass"/>
          <Goal>
	    <xsl:attribute name="goalTag">Invariant is preserved</xsl:attribute>
            <Sub_Calculus>
              <xsl:copy-of select="Body/*"/>
              <Binary_Pred op="=>">
                <Nary_Pred>
                  <xsl:attribute name="op">
                    <xsl:value-of select="$and" />
                  </xsl:attribute>
                  <xsl:for-each select="Includes">
                    <Nary_Pred>
                      <xsl:attribute name="op">
                        <xsl:value-of select="$and" />
                      </xsl:attribute>
		      <xsl:copy-of select="Machine/Invariant/*"/>
		      <xsl:copy-of select="Machine/Assertions/*"/>
		    </Nary_Pred>
		  </xsl:for-each>
		</Nary_Pred>
		<Nary_Pred>
		  <xsl:attribute name="op">
		    <xsl:value-of select="$and" />
		  </xsl:attribute>
		  <xsl:copy-of select="../../Invariant/*"/>
		</Nary_Pred>
	      </Binary_Pred>
            </Sub_Calculus>
          </Goal>
        </Proof_Obligation>
      </xsl:for-each>

      <xsl:for-each select="Sees/Machine/Operations/*">
        <Proof_Obligation ressource="Extended_SEES">
          <Comment>Operations des machines vues dans une machine abstraite</Comment>
          <Tag><xsl:value-of select="@name"/></Tag>
          <Definition name="B definitions"/>
          <Definition name="ctx"/>
          <Definition name="cst"/>
          <Definition name="lprp"/>
          <Definition name="inprp"/>
          <Definition name="inext"/>
          <Definition name="seext"/>
          <Definition name="inv"/>
          <Definition name="ass"/>
          <Goal>
	    <xsl:attribute name="goalTag">Invariant is preserved</xsl:attribute>
            <Sub_Calculus>
              <xsl:copy-of select="Body/*"/>
              <Binary_Pred op="=>">
                <Nary_Pred>
                  <xsl:attribute name="op">
                    <xsl:value-of select="$and" />
                  </xsl:attribute>
                  <xsl:for-each select="Includes">
                    <xsl:copy-of select="Machine/Invariant/*"/>
                    <xsl:copy-of select="Machine/Assertions/*"/>
                  </xsl:for-each>
		</Nary_Pred>
		<Nary_Pred>
		  <xsl:attribute name="op">
		    <xsl:value-of select="$and" />
		  </xsl:attribute>
		  <xsl:copy-of select="/Machine/Invariant/*"/>
		</Nary_Pred>
	      </Binary_Pred>
            </Sub_Calculus>
          </Goal>
        </Proof_Obligation>
      </xsl:for-each>

      <Proof_Obligation>
	<Comment></Comment>
        <Tag>WellDefinednessConstraints</Tag>
	<Goal goalTag="Constraints are well defined">
          <Well_Definedness_Pred>
            <xsl:copy-of select="Constraints/*"/>
          </Well_Definedness_Pred>
        </Goal>
      </Proof_Obligation>

      <Proof_Obligation>
	<Comment></Comment>
        <Tag>WellDefinednessProperties</Tag>
        <Definition name="B definitions"/>
        <Definition name="ctx"/>
        <Definition name="cst"/>
        <Definition name="sets"/>
	<Goal goalTag="Properties are well defined">
          <Well_Definedness_Pred>
            <xsl:copy-of select="Properties/*"/>
          </Well_Definedness_Pred>
        </Goal>
      </Proof_Obligation>

      <Proof_Obligation>
	<Comment></Comment>
        <Tag>WellDefinednessInvariant</Tag>
        <Definition name="B definitions"/>
        <Definition name="ctx"/>
        <Definition name="cst"/>
        <Definition name="lprp"/>
        <Definition name="inprp"/>
        <Definition name="inext"/>
        <Definition name="seext"/>
	<Goal goalTag="Invariant is well defined">
          <Well_Definedness_Pred>
            <xsl:copy-of select="Invariant/*"/>
          </Well_Definedness_Pred>
        </Goal>
      </Proof_Obligation>

      <xsl:for-each select="Assertions/*">
	<xsl:call-template name="machine_WDAssertions">
	  <xsl:with-param name="pos" select="position()" /> 
	</xsl:call-template>
      </xsl:for-each>

      <Proof_Obligation>
	<Comment></Comment>
        <Tag>WellDefinednessInitialisation</Tag>
        <Definition name="B definitions"/>
        <Definition name="ctx"/>
        <Definition name="cst"/>
        <Definition name="lprp"/>
        <Definition name="inprp"/>
        <Definition name="inext"/>
        <Definition name="seext"/>
	<Goal goalTag="Initialisation is well defined">
          <Well_Definedness_Subst>
            <xsl:copy-of select="Initialisation/*"/>
          </Well_Definedness_Subst>
        </Goal>
      </Proof_Obligation>

      <xsl:for-each select="Operations/*">
        <Proof_Obligation>
          <Comment>Operations dans une machine abstraite</Comment>
          <Tag>WellDefinedness_<xsl:value-of select="@name"/></Tag>
          <Definition name="B definitions"/>
          <Definition name="ctx"/>
          <Definition name="cst"/>
          <Definition name="lprp"/>
          <Definition name="inprp"/>
          <Definition name="inext"/>
          <Definition name="seext"/>
          <Definition name="inv"/>
          <Definition name="ass"/>
	  <Goal goalTag="Operation is well defined">
            <Well_Definedness_Subst>
              <xsl:copy-of select="Body/*"/>
            </Well_Definedness_Subst>
          </Goal>
        </Proof_Obligation>
      </xsl:for-each>

      <Proof_Obligation ressource="EventB_DeadlockFreeness">
        <Comment>DLF dans une machine abstraite</Comment>
        <Tag>DLF</Tag>
        <Definition name="B definitions"/>
        <Definition name="ctx"/>
        <Definition name="cst"/>
        <Definition name="lprp"/>
        <Definition name="inprp"/>
        <Definition name="inext"/>
        <Definition name="seext"/>
        <Definition name="inv"/>
        <Definition name="ass"/>
	<Goal>
	  <xsl:attribute name="goalTag">Deadlock Freeness in Abstraction</xsl:attribute>
	  <Nary_Pred op="or">
	    <xsl:for-each select="Operations/Operation">
	      <xsl:apply-templates select="Body/*" mode="guard"/>
	    </xsl:for-each>
	  </Nary_Pred>
	</Goal>
      </Proof_Obligation>

      <xsl:for-each select="Operations/*">
        <Proof_Obligation ressource="Generate_EventB_Feasibility_PO">
          <Comment>Faisabilite des evenements</Comment>
          <Tag>FIS</Tag>
          <Definition name="B definitions"/>
          <Definition name="ctx"/>
          <Definition name="cst"/>
          <Definition name="lprp"/>
          <Definition name="inprp"/>
          <Definition name="inext"/>
          <Definition name="seext"/>
          <Definition name="inv"/>
          <Definition name="ass"/>
	  <Goal>
	    <xsl:attribute name="goalTag">Body of <xsl:value-of select="@name"/> is Feasible </xsl:attribute>
	    <xsl:apply-templates select="Body/*" mode="fis"/>
          </Goal>
        </Proof_Obligation>
      </xsl:for-each>

      <!-- Exclusivite pour les evenements abstraites -->
      <xsl:for-each select="Operations/*">
        <xsl:call-template name="exclusive_events_system_PO">
          <xsl:with-param name="event1" select=".[position()]"/>
          <xsl:with-param name="event_list" select=".[position()]/following-sibling::*"/>
        </xsl:call-template>
      </xsl:for-each>

      <TypeInfos>
        <xsl:copy-of select="TypeInfos/*"/>
      </TypeInfos>
    </Proof_Obligations>
  </xsl:template>

  <xsl:template match="Select" mode="guard">
    <xsl:choose>
      <xsl:when test="count(When_Clauses/*)=1">
        <xsl:copy-of select="When_Clauses/When/Condition/*"/>
      </xsl:when>
      <xsl:otherwise>
        <Exp_Comparison op="=">
          <Integer_Literal value="0" typref="1" />
          <Integer_Literal value="0" typref="1" />
        </Exp_Comparison>
      </xsl:otherwise>
    </xsl:choose>
  </xsl:template>

  <xsl:template match="ANY_Sub" mode="guard">
    <Quantified_Pred type="#">
      <Variables>
        <xsl:for-each select="Variables">
          <xsl:copy-of select="Id[@value]"/>
        </xsl:for-each>
      </Variables>
      <Body>
        <xsl:copy-of select="Pred/*"/>
      </Body>
    </Quantified_Pred>
  </xsl:template>

  <xsl:template match="*[not(name(.)='Select') and not(name(.)='ANY_Sub')]" mode="guard">
    <Exp_Comparison op="=">
      <Integer_Literal value="0" typref="1" />
      <Integer_Literal value="0" typref="1" />
    </Exp_Comparison>
  </xsl:template>

  <xsl:template match="*" mode="fis">
    <Binary_Pred op="=>">
      <xsl:apply-templates select="." mode="guard"/>
      <Not>
        <Sub_Calculus>
          <xsl:copy-of select="."/>
          <Not>
            <Exp_Comparison op="=">
              <Integer_Literal value="0" typref="1" />
              <Integer_Literal value="0" typref="1" />
            </Exp_Comparison>
          </Not>
        </Sub_Calculus>
      </Not>
    </Binary_Pred>
  </xsl:template>

  <xsl:template name="exclusive_events_system_PO">
    <xsl:param name="event1"/>
    <xsl:param name="event_list"/>
    <xsl:for-each select="$event_list">
      <Proof_Obligation ressource="Generate_EventB_Exclusivity_PO">
        <Comment>Exclusivite pour les evenements abstraites</Comment>
        <Tag>EXC</Tag>
        <Definition name="B definitions"/>
        <Definition name="ctx"/>
        <Definition name="cst"/>
        <Definition name="lprp"/>
        <Definition name="inprp"/>
        <Definition name="inext"/>
        <Definition name="seext"/>
        <Definition name="inv"/>
        <Definition name="ass"/>
	<Goal>
	  <xsl:attribute name="goalTag">Exclusivity for Abstract Events not(grd(<xsl:value-of select="$event1/@name"/>) <xsl:value-of select="$and"/>  grd(<xsl:value-of select="./@name"/>))</xsl:attribute>
	  <Binary_Pred op="=>">
	    <Nary_Pred>
	      <xsl:attribute name="op">
		<xsl:value-of select="$and" />
	      </xsl:attribute>
	      <xsl:apply-templates select="$event1/Body/*" mode="guard"/>
	      <xsl:apply-templates select="./Body/*" mode="guard"/>
	    </Nary_Pred>
	    <Exp_Comparison op="=">
	      <Integer_Literal value="0" typref="1" />
	      <Integer_Literal value="1" typref="1" />
	    </Exp_Comparison>
	  </Binary_Pred>
	</Goal>
      </Proof_Obligation>
    </xsl:for-each>
  </xsl:template>

  <xsl:template match="Machine[@type='refinement']">
    <Proof_Obligations>

      <xsl:call-template name="common_part"/>
      <xsl:call-template name="common_part_of_specification"/>
      <xsl:call-template name="common_part_of_refinement"/>
      <xsl:call-template name="common_part2"/>

      <Define name="sets">
        <xsl:copy-of select="Sets/*"/>
      </Define>

      <xsl:for-each select="Includes/*">
        <Proof_Obligation>
          <Comment>Inclusion dans un raffinement</Comment>
          <Tag>InstanciatedConstraintsLemmas</Tag>
          <Definition name="B definitions"/>
          <Definition name="ctx"/>
          <Definition name="mchcst"/>
          <Definition name="aprp"/>
          <Definition name="lprp"/>
	  <Goal goalTag="Instanciation of included machine">
            <Nary_Pred>
              <xsl:attribute name="op">
                <xsl:value-of select="$and" />
              </xsl:attribute>
              <xsl:copy-of select="Constraints/*"/>
            </Nary_Pred>
          </Goal>
        </Proof_Obligation>
      </xsl:for-each>

      <xsl:for-each select="Assertions/*">
        <xsl:call-template name="refinement_Assertions">
          <xsl:with-param name="pos" select="position()"/>
        </xsl:call-template>
      </xsl:for-each>

      <Proof_Obligation>
        <Comment>Initialisation dans un raffinement</Comment>
        <Tag>Initialisation</Tag>
        <Definition name="B definitions"/>
        <Definition name="ctx"/>
        <Definition name="mchcst"/>
        <Definition name="aprp"/>
        <Definition name="lprp"/>
        <Definition name="inprp"/>
        <Definition name="inext"/>
        <Definition name="seext"/>
	<Goal>
	  <xsl:attribute name="goalTag">Refinement is correct</xsl:attribute>
	  <Sub_Calculus>
	    <Nary_Sub op=";">
	      <xsl:copy-of select="Includes/Machine/Initialisation/*"/>
	      <xsl:copy-of select="Initialisation/*"/>
	    </Nary_Sub>
	    <Binary_Pred op="=>">
	      <Nary_Pred>
		<xsl:attribute name="op">
		  <xsl:value-of select="$and" />
		</xsl:attribute>
		<xsl:copy-of select="Includes/Machine/Invariant/*"/>
		<xsl:copy-of select="Includes/Machine/Assertions/*"/>
	      </Nary_Pred>
	      <Not>
		<Sub_Calculus>
		  <Nary_Sub op=";">
		    <xsl:copy-of select="/Machine/Abstraction/Machine[1]/Includes/Machine/Initialisation/*"/>
		    <xsl:copy-of select="/Machine/Abstraction/Machine[1]/Initialisation/*"/>
		  </Nary_Sub>
		  <Not>
		    <Nary_Pred>
		      <xsl:attribute name="op">
			<xsl:value-of select="$and" />
		      </xsl:attribute>
		      <xsl:copy-of select="/Machine/Invariant/*"/>
		    </Nary_Pred>
		  </Not>
		</Sub_Calculus>
	      </Not>
	    </Binary_Pred>
	  </Sub_Calculus>
	</Goal>
      </Proof_Obligation>

      <xsl:for-each select="Operations/Operation|Operations/ImplicitOperation">
        <xsl:choose>
          <xsl:when test = "boolean(Refines)">
            <xsl:choose>
              <xsl:when test="count(Refines/*)=1">
                <xsl:call-template name="Operation_refinement">
                  <xsl:with-param name="opName" select="@name"/>
                  <xsl:with-param name="abstractOpName" select="Refines/Id/@value"/>
                </xsl:call-template>
              </xsl:when>
              <xsl:otherwise>
                <xsl:call-template name="refinement_merge">
                  <xsl:with-param name="opName" select="@name"/>
                  <xsl:with-param name="refined_events" select="Refines/*"/>
                </xsl:call-template>
              </xsl:otherwise>
            </xsl:choose>
          </xsl:when>
          <xsl:otherwise>
            <xsl:call-template name="Operation_refinement">
              <xsl:with-param name="opName" select="@name"/>
            </xsl:call-template>
          </xsl:otherwise>
        </xsl:choose>
      </xsl:for-each>

      <xsl:for-each select="Sees/Machine/Operations/*">
        <Proof_Obligation ressource="Extended_SEES">
          <Comment>Operations des machines vues dans un raffinement</Comment>
          <Tag><xsl:value-of select="@name"/></Tag>
          <Definition name="B definitions"/>
          <Definition name="ctx"/>
          <Definition name="mchcst"/>
          <Definition name="aprp"/>
          <Definition name="lprp"/>
          <Definition name="inprp"/>
          <Definition name="inext"/>
          <Definition name="seext"/>
          <Definition name="abs"/>
          <Definition name="inv"/>
          <Definition name="ass"/>
          <Goal>
	    <xsl:attribute name="goalTag">Invariant is preserved</xsl:attribute>
            <Sub_Calculus>
              <xsl:copy-of select="Body/*"/>
              <Binary_Pred op="=>">
                <Nary_Pred>
                  <xsl:attribute name="op">
                    <xsl:value-of select="$and" />
                  </xsl:attribute>
                  <xsl:copy-of select="Includes/Machine/Invariant/*"/>
                  <xsl:copy-of select="Includes/Machine/Assertions/*"/>
		</Nary_Pred>
		<Nary_Pred>
		  <xsl:attribute name="op">
		    <xsl:value-of select="$and" />
		  </xsl:attribute>
		  <xsl:copy-of select="/Machine/Invariant/*"/>
		</Nary_Pred>
	      </Binary_Pred>
            </Sub_Calculus>
          </Goal>
        </Proof_Obligation>
      </xsl:for-each>

      <Proof_Obligation>
	<Comment></Comment>
        <Tag>WellDefinednessProperties</Tag>
        <Definition name="B definitions"/>
        <Definition name="ctx"/>
        <Definition name="mchcst"/>
        <Definition name="aprp"/>
        <Definition name="sets"/>
	<Goal goalTag="Properties are well defined">
          <Well_Definedness_Pred>
            <xsl:copy-of select="Properties/*"/>
          </Well_Definedness_Pred>
        </Goal>
      </Proof_Obligation>

      <Proof_Obligation>
	<Comment></Comment>
        <Tag>WellDefinednessInvariant</Tag>
        <Definition name="B definitions"/>
        <Definition name="ctx"/>
        <Definition name="mchcst"/>
        <Definition name="aprp"/>
        <Definition name="lprp"/>
        <Definition name="inprp"/>
        <Definition name="inext"/>
        <Definition name="seext"/>
        <Definition name="abs"/>
	<Goal goalTag="Invariant is well defined">
          <Well_Definedness_Pred>
            <xsl:copy-of select="Invariant/*"/>
          </Well_Definedness_Pred>
        </Goal>
      </Proof_Obligation>

      <xsl:for-each select="Assertions/*">
	<xsl:call-template name="refinement_WDAssertions">
	  <xsl:with-param name="pos" select="position()"/> 
	</xsl:call-template>
      </xsl:for-each>

      <Proof_Obligation>
	<Comment></Comment>
        <Tag>WellDefinednessInitialisation</Tag>
        <Definition name="B definitions"/>
        <Definition name="ctx"/>
        <Definition name="mchcst"/>
        <Definition name="aprp"/>
        <Definition name="lprp"/>
        <Definition name="inprp"/>
        <Definition name="inext"/>
        <Definition name="seext"/>
	<Goal goalTag="Initialisation is well defined">
          <Well_Definedness_Subst>
            <xsl:copy-of select="Initialisation/*"/>
          </Well_Definedness_Subst>
        </Goal>
      </Proof_Obligation>

      <xsl:for-each select="Operations/*">
        <Proof_Obligation>
          <Comment>Operations dans un raffinement</Comment>
          <Tag>WellDefinedness_<xsl:value-of select="@name"/></Tag>
          <Definition name="B definitions"/>
          <Definition name="ctx"/>
          <Definition name="mchcst"/>
          <Definition name="aprp"/>
          <Definition name="lprp"/>
          <Definition name="inprp"/>
          <Definition name="inext"/>
          <Definition name="seext"/>
          <Definition name="abs"/>
          <Definition name="inv"/>
          <Definition name="ass"/>
	  <Goal goalTag="Operation precondition is well defined">
            <Well_Definedness_Subst>
              <xsl:copy-of select="Body/*"/>
            </Well_Definedness_Subst>
          </Goal>
        </Proof_Obligation>
      </xsl:for-each>


      <!-- Absence relative de blockage pour les evenements raffines -->
      <xsl:for-each select="/Machine/Abstraction/Machine[1]/Operations/*">
        <xsl:if test="not(clearsy:is_closed_event(.))">
          <Proof_Obligation ressource="EventB_DeadlockFreeness">
            <Comment>DLF dans un raffinement</Comment>
            <Tag>DLF</Tag>
            <Definition name="B definitions"/>
            <Definition name="ctx"/>
            <Definition name="mchcst"/>
            <Definition name="aprp"/>
            <Definition name="lprp"/>
            <Definition name="inprp"/>
            <Definition name="inext"/>
            <Definition name="seext"/>
            <Definition name="abs"/>
            <Definition name="inv"/>
            <Definition name="ass"/>
            <Goal>
	      <xsl:attribute name="goalTag">Deadlock Freeness in Refinement for Event <xsl:copy-of select="@name"/></xsl:attribute>
	      <Binary_Pred op="=>">
		<xsl:apply-templates select="Body/*" mode="guard"/>
		<Nary_Pred op="or">
		  <xsl:for-each select="/Machine/Operations/*">
		    <xsl:if test="not(clearsy:is_closed_event(.))">
		      <xsl:apply-templates select="current()/Body/*" mode="guard"/>
		    </xsl:if>
		  </xsl:for-each>
		</Nary_Pred>
	      </Binary_Pred>
            </Goal>
          </Proof_Obligation>
        </xsl:if>
      </xsl:for-each>
      
      <!-- Faisabilit pour les evenements raffines -->
      <xsl:for-each select="Operations/Operation">
        <xsl:if test="not(clearsy:is_closed_event(.))">
          <Proof_Obligation ressource="Generate_EventB_Feasibility_PO">
            <Comment>Faisabilite des evenements</Comment>
            <Tag>FIS</Tag>
            <Definition name="B definitions"/>
            <Definition name="ctx"/>
            <Definition name="mchcst"/>
            <Definition name="aprp"/>
            <Definition name="lprp"/>
            <Definition name="inprp"/>
            <Definition name="inext"/>
            <Definition name="seext"/>
            <Definition name="abs"/>
            <Definition name="inv"/>
            <Definition name="ass"/>
            <Goal>
	      <xsl:attribute name="goalTag">Body of <xsl:value-of select="@name"/> is feasible </xsl:attribute>
	      <xsl:apply-templates select="Body/*" mode="fis"/>
            </Goal>
          </Proof_Obligation>
        </xsl:if>
      </xsl:for-each>
      
      <!-- Non divergence pour les evenements raffines. Le variant est un naturel -->
      <xsl:if test="/Machine/Variant">
        <Proof_Obligation ressource="Generate_EventB_Non_Divergence_PO">
          <Comment>Le variant est un naturel</Comment>
          <Tag>NAT</Tag>
          <Definition name="B definitions"/>
          <Definition name="ctx"/>
          <Definition name="mchcst"/>
          <Definition name="aprp"/>
          <Definition name="lprp"/>
          <Definition name="inprp"/>
          <Definition name="inext"/>
          <Definition name="seext"/>
          <Definition name="abs"/>
          <Definition name="inv"/>
          <Definition name="ass"/>
	  <Goal>
	    <xsl:attribute name="goalTag">Variant is a NATURAL</xsl:attribute>
	    <Exp_Comparison op=":">
	      <xsl:copy-of select="/Machine/Variant/*"/>
              <Id value="NATURAL" typref="2"/>
	    </Exp_Comparison>
	  </Goal>
        </Proof_Obligation>
      </xsl:if>

       <!-- Non divergence pour les evenements raffines. Le nouveaux evenement decroissent le variant -->
      <xsl:if test="/Machine/Variant">
        <xsl:for-each select="$new_events_variant/events/*">
          <Proof_Obligation ressource="Generate_EventB_Non_Divergence_PO">
            <Comment>Non divergence des nouveaux evenements</Comment>
            <Tag>NDI</Tag>
            <Definition name="B definitions"/>
            <Definition name="ctx"/>
            <Definition name="mchcst"/>
            <Definition name="aprp"/>
            <Definition name="lprp"/>
            <Definition name="inprp"/>
            <Definition name="inext"/>
            <Definition name="seext"/>
            <Definition name="abs"/>
            <Definition name="inv"/>
	    <Definition name="ass"/>
	    <Goal>
	      <xsl:attribute name="goalTag">Non-divergence of new event <xsl:value-of select="@name"/></xsl:attribute>
	      <xsl:call-template name="po_variant">
		<xsl:with-param name="variant" select="$new_events_variant/variant/*"/>
	      </xsl:call-template>
	    </Goal>
          </Proof_Obligation>
        </xsl:for-each>
      </xsl:if>

      <!--Couverture pour les evenements abstraits. -->
      <xsl:for-each select="$abstract_concrete_map/map/*">
        <xsl:if test="not(./key/*/@name=./value/ImplicitOperation/@name)"> <!-- TRUE if abstract event is  modified in refinement -->
          <Proof_Obligation ressource="Generate_EventB_Coverage_PO">
            <Comment>Coverage pour les evenements abstraits</Comment>
            <Tag>COV</Tag>
            <Definition name="B definitions"/>
            <Definition name="ctx"/>
            <Definition name="mchcst"/>
            <Definition name="aprp"/>
            <Definition name="lprp"/>
            <Definition name="inprp"/>
            <Definition name="inext"/>
            <Definition name="seext"/>
            <Definition name="abs"/>
            <Definition name="inv"/>
            <Definition name="ass"/>
	    <Goal>
	      <xsl:attribute name="goalTag">Coverage for Abstract Event <xsl:value-of select="./key/*/@name"/></xsl:attribute>
	      <Binary_Pred op="=>">
		<xsl:apply-templates select="./key/*/Body/*" mode="guard"/>
		<Nary_Pred op="or">
		  <xsl:for-each select="./value/*/Body">
		    <xsl:apply-templates select="./*" mode="guard"/>
		  </xsl:for-each>
		  <xsl:for-each select="$new_events_variant/events/*/Body">
		    <xsl:apply-templates select="./*" mode="guard"/>
		  </xsl:for-each>
		</Nary_Pred>
	      </Binary_Pred>
	    </Goal>
          </Proof_Obligation>
        </xsl:if>
      </xsl:for-each>

      <!-- Exclusivite pour les evenements raffines. Evenements raffinant un evenement -->
      <xsl:for-each select="$abstract_concrete_map/map/*">
        <xsl:call-template name="exclusive_refinement_one">
          <xsl:with-param name="event_list" select="./value/*"/>
        </xsl:call-template>
      </xsl:for-each>

      <!-- Exclusivite pour les evenements raffines. Evenements raffines vs nouveaux evenements -->
      <xsl:for-each select="/Machine/Operations/*">
        <xsl:call-template name="exclusive_events_refinement_new_PO">
          <xsl:with-param name="event1" select="."/>
        </xsl:call-template>
      </xsl:for-each>

      <!-- Exclusivite pour les evenements raffines. Nouveaux evenements -->
      <xsl:for-each select="$new_events_variant/events/*">
        <xsl:call-template name="exclusive_new_events_PO">
          <xsl:with-param name="event1" select=".[position()]"/>
          <xsl:with-param name="event_list" select=".[position()]/following-sibling::*"/>
        </xsl:call-template>
      </xsl:for-each>

      <TypeInfos>
        <xsl:copy-of select="TypeInfos/*"/>
      </TypeInfos>
    </Proof_Obligations>
    
  </xsl:template>

  <xsl:template name="exclusive_refinement_one">
    <xsl:param name="event_list"/>
    <xsl:for-each select="$event_list">
      <xsl:call-template name="exclusive_events_refinement_PO">
        <xsl:with-param name="event1" select=".[position()]"/>
        <xsl:with-param name="event_list" select=".[position()]/following-sibling::*"/>
      </xsl:call-template>
    </xsl:for-each>
  </xsl:template>

  <xsl:template name="exclusive_new_events_PO">
    <xsl:param name="event1"/>
    <xsl:param name="event_list"/>
    <xsl:for-each select="$event_list">
      <Proof_Obligation ressource="Generate_EventB_Exclusivity_PO">
        <Comment>Exclusivite pour les  nouveaux evenements</Comment>
        <Tag>EXC</Tag>
        <Definition name="B definitions"/>
        <Definition name="ctx"/>
        <Definition name="mchcst"/>
        <Definition name="aprp"/>
        <Definition name="lprp"/>
        <Definition name="inprp"/>
        <Definition name="inext"/>
        <Definition name="seext"/>
        <Definition name="abs"/>
        <Definition name="inv"/>
        <Definition name="ass"/>
	<Goal>
	  <xsl:attribute name="goalTag">Exclusivity for New Events not(grd(<xsl:value-of select="$event1/@name"/>) <xsl:value-of select="$and"/>  grd(<xsl:value-of select="./@name"/>))</xsl:attribute>
	  <Binary_Pred op="=>">
	    <Nary_Pred>
	      <xsl:attribute name="op">
		<xsl:value-of select="$and" />
	      </xsl:attribute>
	      <xsl:apply-templates select="$event1/Body/*" mode="guard"/>
	      <xsl:apply-templates select="./Body/*" mode="guard"/>
	    </Nary_Pred>
	    <Exp_Comparison op="=">
	      <Integer_Literal value="0" typref="1" />
	      <Integer_Literal value="1" typref="1" />
	    </Exp_Comparison>
	  </Binary_Pred>
	</Goal>
      </Proof_Obligation>
    </xsl:for-each>
  </xsl:template>

  <xsl:template name="exclusive_events_refinement_new_PO">
    <xsl:param name="event1"/>
    <xsl:for-each select="$new_events_variant/events/*">
      <xsl:if test="count($new_events_variant/events/*[@name=$event1/@name])=0"> <!-- Event in $event1 is not new -->
        <Proof_Obligation ressource="Generate_EventB_Exclusivity_PO">
          <Comment>Exclusivite pour les evenements raffines. Cas evenements raffinnes vs nouveaux evenements</Comment>
          <Tag>EXC</Tag>
          <Definition name="B definitions"/>
          <Definition name="ctx"/>
          <Definition name="mchcst"/>
          <Definition name="aprp"/>
          <Definition name="lprp"/>
          <Definition name="inprp"/>
          <Definition name="inext"/>
          <Definition name="seext"/>
          <Definition name="abs"/>
          <Definition name="inv"/>
          <Definition name="ass"/>
          <Goal>
	    <xsl:attribute name="goalTag">Exclusivity for Refined and New Events not(grd(<xsl:value-of select="$event1/@name"/>) <xsl:value-of select="$and"/>  grd(<xsl:value-of select="./@name"/>))</xsl:attribute>
	    <Binary_Pred op="=>">
	      <Nary_Pred>
		<xsl:attribute name="op">
		  <xsl:value-of select="$and" />
		</xsl:attribute>
		<xsl:apply-templates select="$event1/Body/*" mode="guard"/>
		<xsl:apply-templates select="./Body/*" mode="guard"/>
	      </Nary_Pred>
	      <Exp_Comparison op="=">
		<Integer_Literal value="0" typref="1" />
		<Integer_Literal value="1" typref="1" />
	      </Exp_Comparison>
	    </Binary_Pred>
	  </Goal>
        </Proof_Obligation>
      </xsl:if>
    </xsl:for-each>
  </xsl:template>

  <xsl:template name="exclusive_events_refinement_PO">
    <xsl:param name="event1"/>
    <xsl:param name="event_list"/>
    <xsl:for-each select="$event_list">
      <Proof_Obligation ressource="Generate_EventB_Exclusivity_PO">
        <Comment>Exclusivite pour les evenements raffines. Cas evenements raffinant un evenement</Comment>
        <Tag>EXC</Tag>
        <Definition name="B definitions"/>
        <Definition name="ctx"/>
        <Definition name="mchcst"/>
        <Definition name="aprp"/>
        <Definition name="lprp"/>
        <Definition name="inprp"/>
        <Definition name="inext"/>
        <Definition name="seext"/>
        <Definition name="abs"/>
        <Definition name="inv"/>
	<Definition name="ass"/>
	<Goal>
	  <xsl:attribute name="goalTag">Exclusivity for Events Refining the Same Event not(grd(<xsl:value-of select="$event1/@name"/>) <xsl:value-of select="$and"/>  grd(<xsl:value-of select="./@name"/>))</xsl:attribute>
	  <Binary_Pred op="=>">
	    <Nary_Pred>
	      <xsl:attribute name="op">
		<xsl:value-of select="$and" />
	      </xsl:attribute>
	      <xsl:apply-templates select="$event1/Body/*" mode="guard"/>
	      <xsl:apply-templates select="./Body/*" mode="guard"/>
	    </Nary_Pred>
	    <Exp_Comparison op="=">
	      <Integer_Literal value="0" typref="1" />
	      <Integer_Literal value="1" typref="1" />
	    </Exp_Comparison>
	  </Binary_Pred>
	</Goal>
      </Proof_Obligation>
    </xsl:for-each>
  </xsl:template>

  <xsl:variable name="abstract_concrete_map">
    <map>
      <xsl:for-each select="/Machine/Abstraction/Machine[1]/Operations/*">
        <xsl:if test="not(clearsy:is_closed_event(.))">
        <map_entry>
          <key>
            <xsl:copy-of select="."/> 
          </key>
          <xsl:call-template name="fill_value_map">
            <xsl:with-param name="abstract_event_name" select="./@name"/>
          </xsl:call-template>
        </map_entry>
        </xsl:if>
      </xsl:for-each>
    </map>
  </xsl:variable>
  
  <!-- only SELECT 0=1 THEN skip END or SELECT 1=0 skip END are considered as syntaxically closed -->
  <!-- any other event with guard x=y where x is not 0 or 1 is not considered as syntaxically closed -->
  <!-- an implicitly closed event has guard x=y with x and y different -->
  <!-- implicit closed event generate obvious POs -->
  
  <xsl:function name="clearsy:is_closed_event">
    <xsl:param name="event"/>

    <xsl:if test="not($event/Body/Select)">
      <xsl:sequence select="false()"/>
    </xsl:if> <!-- end not($event/Body/Select) -->
    <xsl:if test="$event/Body/Select">
      <xsl:if test="not(count($event/Body/Select/When_Clauses/When)=1)">
        <xsl:sequence select="false()"/> <!-- return value false -->
      </xsl:if> <!-- end not(count($event/Body/Select/When_Clauses/When)=1)-->

      <xsl:if test="count($event/Body/Select/When_Clauses/When)=1">
        
        <xsl:if test="not($event/Body/Select/When_Clauses/When/Then/Skip)">
          <xsl:sequence select="false()"/>
        </xsl:if> <!--end not($event/Body/Select/When_Clauses/When/Then/Skip) -->

        <xsl:if test="$event/Body/Select/When_Clauses/When/Then/Skip">
          <xsl:if test="not(count($event/Body/Select/When_Clauses/When/Condition/Exp_Coparison/@op='=')=1)">
            <xsl:sequence select="false()"/>
          </xsl:if>   <!--end not($event/Body/Select/When_Clauses/When/Condition/Exp_Coparison/@op='='-->

          <xsl:if test="count($event/Body/Select/When_Clauses/When/Condition/Exp_Coparison/@op='=')=1">
            
            <xsl:choose>
              <xsl:when test="$event/Body/Select/When_Clauses/When/Condition/Exp_Comparison/Integer_Literal[1]/@value=0">
                <xsl:choose>
                  <xsl:when test="$event/Body/Select/When_Clauses/When/Condition/Exp_Comparison/Integer_Literal[2]/@value=1">
                    <xsl:sequence select="true()"/> <!-- return value true -->
                  </xsl:when>
                  <xsl:otherwise>
                    <xsl:sequence select="false()"/>
                  </xsl:otherwise>
                </xsl:choose>
              </xsl:when>
              <xsl:when test="$event/Body/Select/When_Clauses/When/Condition/Exp_Comparison/Integer_Literal[1]/@value=1">
                <xsl:choose>
                  <xsl:when test="$event/Body/Select/When_Clauses/When/Condition/Exp_Comparison/Integer_Literal[2]/@value=0">
                    <xsl:sequence select="true()"/>
                  </xsl:when>
                  <xsl:otherwise>
                    <xsl:sequence select="false()"/>
                  </xsl:otherwise>
                </xsl:choose>
              </xsl:when>
              <xsl:otherwise>
                <xsl:sequence select="false()"/>
              </xsl:otherwise>
            </xsl:choose>

          </xsl:if> <!--end Exp_Coparison/@op='=' -->
          
        </xsl:if>  <!-- end Skip -->
      </xsl:if> <!-- end count($event/Body/Select/When_Clauses/When)=1 -->
      
    </xsl:if> <!--end $event/Body/Select-->    
    
  </xsl:function>

  <xsl:template name="fill_value_map">
    <xsl:param name="abstract_event_name"/>
    <value>
      <xsl:for-each select="/Machine/Operations/*">
        <xsl:if test="./@name=$abstract_event_name">
          <xsl:if test="not(clearsy:is_closed_event(.))">
            <xsl:copy-of select="."/>
          </xsl:if>
        </xsl:if>
        <xsl:if test="./Refines/Id/@value=$abstract_event_name">
          <xsl:copy-of select="."/>
        </xsl:if>
      </xsl:for-each>
    </value>
  </xsl:template>

  <xsl:variable name="new_events_variant">
    <events>
      <xsl:for-each select="/Machine/Operations/*">
        <xsl:if test="not(Refines)">
          <xsl:if test="count(/Machine/Abstraction/Machine[1]/Operations/*[@name=current()/@name])=0">
            <xsl:copy-of select="."/>
          </xsl:if>
        </xsl:if>
      </xsl:for-each>
    </events>
    <variant>
      <xsl:copy-of select="/Machine/Variant"/>
    </variant>
  </xsl:variable>

  <xsl:variable name="lowerthan"><![CDATA[<i]]></xsl:variable>

  <xsl:template name="po_variant">
    <xsl:param name="variant"/>
    <Let_Fresh_Id name="xx">
      <Sub_Calculus>
        <Simple_Assignement_Sub>
          <Variables>
            <Fresh_Id ref="xx" typref="1"/>
          </Variables>
          <Values>
            <xsl:copy-of select="$variant/*"/>
          </Values>
        </Simple_Assignement_Sub>
        <Sub_Calculus>
          <xsl:copy-of select="Body/*"/>
          <Exp_Comparison>
            <xsl:attribute name="op">
              <xsl:value-of select="$lowerthan" />
            </xsl:attribute>
            <xsl:copy-of select="$variant/*"/>
            <Fresh_Id ref="xx" typref="1"/>
          </Exp_Comparison>
        </Sub_Calculus>
      </Sub_Calculus>
    </Let_Fresh_Id>
  </xsl:template>

  <xsl:template match="Machine[@type='implementation']">
    <Proof_Obligations>

      <xsl:call-template name="common_part"/>
      <xsl:call-template name="common_part_of_refinement"/>
      <xsl:call-template name="common_part2"/>

      <Define name="imlprp">
        <xsl:copy-of select="Sets/*"/>
        <xsl:copy-of select="Properties/*"/>
        <xsl:for-each select="Values/Valuation/*">
          <Exp_Comparison op="=">
            <Id>
              <xsl:attribute name="value">
                <xsl:copy-of select="../@ident"/>
              </xsl:attribute>
              <xsl:copy-of select="./Attr"/>
            </Id>
            <xsl:copy-of select="."/>
          </Exp_Comparison>
        </xsl:for-each>
      </Define>

      <Define name="imprp">
        <xsl:copy-of select="Imports/Machine/Sets/*"/>
        <xsl:copy-of select="Imports/Machine/Includes/Machine/Sets/*"/>
        <xsl:copy-of select="Imports/Machine/Properties/*"/>
        <xsl:copy-of select="Imports/Machine/Includes/Machine/Properties/*"/>
      </Define>

      <Define name="imext">
        <xsl:copy-of select="Imports/Machine/Invariant/*"/>
        <xsl:copy-of select="Imports/Machine/Assertions/*"/>
        <xsl:copy-of select="Imports/Machine/Includes/Machine/Invariant/*"/>
        <xsl:copy-of select="Imports/Machine/Includes/Machine/Assertions/*"/>
      </Define>

      <Define name="sets">
        <xsl:copy-of select="Sets/*"/>
      </Define>

      <xsl:for-each select="Imports/*">
        <Proof_Obligation>
          <Comment>Importation dans une implantation</Comment>
          <Tag>InstanciatedConstraintsLemmas</Tag>
          <Definition name="B definitions"/>
          <Definition name="ctx"/>
          <Definition name="mchcst"/>
          <Definition name="aprp"/>
          <Definition name="imlprp"/>
	  <Goal goalTag="Instanciation of imported machine">
            <Nary_Pred>
              <xsl:attribute name="op">
                <xsl:value-of select="$and" />
              </xsl:attribute>
              <xsl:copy-of select="Constraints/*"/>
            </Nary_Pred>
          </Goal>
        </Proof_Obligation>
      </xsl:for-each>

      <Proof_Obligation>
        <Comment>Valuation dans une implantation</Comment>
        <Tag>ValuesLemmas</Tag>
        <Definition name="B definitions"/>
        <Definition name="ctx"/>
        <Definition name="imprp"/>
	<Goal>
	  <xsl:attribute name="goalTag">Valuation is correct</xsl:attribute>
	  <Quantified_Pred type="#">
	    <Variables>
	      <xsl:copy-of select="Abstraction[1]/Machine/Abstract_Constants/*"/>
	    </Variables>
	    <Body>
	      <Sub_Calculus>
		<Nary_Sub op=";">
		  <xsl:for-each select="Values/Valuation/*">
		    <Simple_Assignement_Sub>
		      <Variables>
			<Id>
			  <xsl:attribute name="value">
			    <xsl:copy-of select="../@ident"/>
			  </xsl:attribute>
			</Id>
		      </Variables>
		      <Values>
			<xsl:copy-of select="."/>
		      </Values>
		    </Simple_Assignement_Sub>
		  </xsl:for-each>
		</Nary_Sub>
		<Nary_Pred>
		  <xsl:attribute name="op">
		    <xsl:value-of select="$and" />
		  </xsl:attribute>
		  <xsl:copy-of select="Imports/Machine/Properties/*"/>
		  <xsl:copy-of select="Abstraction/Machine/Properties/*"/>
		  <xsl:copy-of select="Abstraction/Machine/Includes/Machine/Properties/*"/>
		  <xsl:copy-of select="Properties/*"/>
                  <xsl:for-each select="Abstraction/Machine/Sets/Set[not(Enumerated_Values)]|Abstraction/Machine/Includes/Machine/Sets/Set[not(Enumerated_Values)]">
                    <Exp_Comparison op=":">
                      <xsl:copy-of select="./Id"/>
                      <Unary_Exp op="FIN" typref="3">
                        <Id value="INTEGER" typref="2"/>
                      </Unary_Exp>
                    </Exp_Comparison>
                    <Unary_Pred op="not">
                      <Exp_Comparison op="=">
                        <xsl:copy-of select="./Id"/>
                        <EmptySet>
                          <xsl:attribute name="typref">
                            <xsl:value-of select="./Id/@typref" />
                          </xsl:attribute>
                        </EmptySet>
                      </Exp_Comparison>
                    </Unary_Pred>
                  </xsl:for-each>
                </Nary_Pred>
	      </Sub_Calculus>
	    </Body>
	  </Quantified_Pred>
	</Goal>
      </Proof_Obligation>

      <xsl:for-each select="Assertions/*">
	<xsl:call-template name="implantation_Assertions">
	  <xsl:with-param name="pos" select="position()"/> 
	</xsl:call-template>
      </xsl:for-each>

      <Proof_Obligation>
          <Comment>Initialisation dans une implantation</Comment>
          <Tag>Initialisation</Tag>
          <Definition name="B definitions"/>
          <Definition name="ctx"/>
          <Definition name="mchcst"/>
          <Definition name="aprp"/>
          <Definition name="imlprp"/>
          <Definition name="imprp"/>
          <Definition name="imext"/>
          <Definition name="seext"/>
	  <Goal>
	    <xsl:attribute name="goalTag">Refinement is correct</xsl:attribute>
	    <Sub_Calculus overflow="true">
	      <Nary_Sub op=";">
		<xsl:copy-of select="Imports/Machine/Initialisation/*"/>
		<xsl:copy-of select="Initialisation/*"/>
	      </Nary_Sub>
	      <Binary_Pred op="=>">
		<Nary_Pred>
		  <xsl:attribute name="op">
		    <xsl:value-of select="$and" />
		  </xsl:attribute>
		  <xsl:copy-of select="Imports/Machine/Invariant/*"/>
		  <xsl:copy-of select="Imports/Machine/Assertions/*"/>
		</Nary_Pred>
		<Not>
		  <Sub_Calculus>
		    <Nary_Sub op=";">
		      <xsl:copy-of select="/Machine/Abstraction/Machine[1]/Includes/Machine/Initialisation/*"/>
		      <xsl:copy-of select="/Machine/Abstraction/Machine[1]/Initialisation/*"/>
		    </Nary_Sub>
		    <Not>
		      <Nary_Pred>
			<xsl:attribute name="op">
			  <xsl:value-of select="$and" />
			</xsl:attribute>
			<xsl:copy-of select="/Machine/Invariant/*"/>
		      </Nary_Pred>
		    </Not>
		  </Sub_Calculus>
		</Not>
	      </Binary_Pred>
	    </Sub_Calculus>
	  </Goal>
      </Proof_Obligation>

      <xsl:for-each select="Local_Operations/Operation">
        <Proof_Obligation>
          <Comment>Specification d’operation locale dans une implantation</Comment>
          <Tag><xsl:value-of select="@name"/></Tag>
          <Definition name="B definitions"/>
          <Definition name="ctx"/>
          <Definition name="mchcst"/>
          <Definition name="aprp"/>
          <Definition name="imlprp"/>
          <Definition name="imprp"/>
          <Definition name="seext"/>
          <Hypothesis>
            <Nary_Pred>
              <xsl:attribute name="op">
                <xsl:value-of select="$and" />
              </xsl:attribute>
              <xsl:copy-of select="/Machine/Imports/Machine/Unrenamed_Invariant/*"/>
            </Nary_Pred>
          </Hypothesis>
          <Hypothesis><xsl:copy-of select="Precondition/*"/></Hypothesis>
          <!-- WARNING : il manque les prédicats de typage de variables concretes de l'implémentation et des machines importées -->
          <Goal>
	    <xsl:attribute name="goalTag">Imported machine invariant is preserved</xsl:attribute>
            <Sub_Calculus>
              <xsl:attribute name="overflow">
                <xsl:value-of select="'false'" />
              </xsl:attribute>
	      <xsl:copy-of select="Body/*"/>
	      <Nary_Pred>
		<xsl:attribute name="op">
		  <xsl:value-of select="$and" />
		</xsl:attribute>
		<xsl:copy-of select="/Machine/Imports/Machine/Unrenamed_Invariant/*"/>
	      </Nary_Pred>
	    </Sub_Calculus>
          </Goal>
        </Proof_Obligation>
      </xsl:for-each>

      <xsl:for-each select="Operations/Operation">
        <xsl:choose>
          <xsl:when test = "boolean(Refines)">
            <xsl:for-each select="Refines/*">
              <xsl:call-template name="Operation_implantation">
                <xsl:with-param name="opName" select="../../@name"/>
                <xsl:with-param name="abstractOpName" select="@value"/>
              </xsl:call-template>
            </xsl:for-each>
          </xsl:when>
          <xsl:otherwise>
            <xsl:call-template name="Operation_implantation">
              <xsl:with-param name="opName" select="@name"/>
            </xsl:call-template>
          </xsl:otherwise>
        </xsl:choose>
      </xsl:for-each>

      <xsl:for-each select="/Machine/Abstraction/Machine[1]/Operations/Operation">
        <xsl:call-template name="Operation_implantation">
          <xsl:with-param name="opName" select="@name"/>
        </xsl:call-template>
      </xsl:for-each>

      <xsl:for-each select="/Machine/Abstraction/Machine[1]/Operations/ImplicitOperation">
        <xsl:call-template name="Operation_implantation">
          <xsl:with-param name="opName" select="@name"/>
        </xsl:call-template>
      </xsl:for-each>

      <xsl:for-each select="/Machine/Local_Operations/Operation">
        <xsl:call-template name="local_implantation">
          <xsl:with-param name="opName" select="@name"/>
        </xsl:call-template>
      </xsl:for-each>

      <xsl:for-each select="Sees/Machine/Operations/*">
        <Proof_Obligation ressource="Extended_SEES">
          <Comment>Operations des machines vues dans une implantation</Comment>
          <Tag><xsl:value-of select="@name"/></Tag>
          <Definition name="B definitions"/>
          <Definition name="ctx"/>
          <Definition name="mchcst"/>
          <Definition name="aprp"/>
          <Definition name="imlprp"/>
          <Definition name="imprp"/>
          <Definition name="imext"/>
          <Definition name="seext"/>
          <Definition name="abs"/>
          <Definition name="inv"/>
          <Definition name="ass"/>
          <Hypothesis><xsl:copy-of select="Precondition/*"/></Hypothesis>
          <Goal>
	    <xsl:attribute name="goalTag">Invariant is preserved</xsl:attribute>
            <Sub_Calculus>
              <xsl:copy-of select="Body/*"/>
                <Binary_Pred op="=>">
                <Nary_Pred>
                  <xsl:attribute name="op">
                    <xsl:value-of select="$and" />
                    </xsl:attribute>
                  <xsl:for-each select="Includes">
                    <xsl:copy-of select="Machine/Invariant/*"/>
                    <xsl:copy-of select="Machine/Assertions/*"/>
                  </xsl:for-each>
		</Nary_Pred>
		<Nary_Pred>
		  <xsl:attribute name="op">
		    <xsl:value-of select="$and" />
		  </xsl:attribute>
		  <xsl:copy-of select="/Machine/Invariant/*"/>
		</Nary_Pred>
	      </Binary_Pred>
            </Sub_Calculus>
          </Goal>
        </Proof_Obligation>
      </xsl:for-each>

      <Proof_Obligation>
	<Comment></Comment>
        <Tag>WellDefinednessProperties</Tag>
        <Definition name="B definitions"/>
        <Definition name="ctx"/>
        <Definition name="mchcst"/>
        <Definition name="aprp"/>
        <Definition name="imprp"/>
        <Definition name="sets"/>
        <Goal goalTag="Properties are well defined">
          <Well_Definedness_Pred>
            <xsl:copy-of select="Properties/*"/>
          </Well_Definedness_Pred>
        </Goal>
      </Proof_Obligation>

      <Proof_Obligation>
	<Comment></Comment>
        <Tag>WellDefinednessInvariant</Tag>
        <Definition name="B definitions"/>
        <Definition name="ctx"/>
        <Definition name="mchcst"/>
        <Definition name="aprp"/>
        <Definition name="imlprp"/>
        <Definition name="imprp"/>
        <Definition name="imext"/>
        <Definition name="seext"/>
        <Definition name="abs"/>
	<Goal goalTag="Invariant is well defined">
          <Well_Definedness_Pred>
            <xsl:copy-of select="Invariant/*"/>
          </Well_Definedness_Pred>
        </Goal>
      </Proof_Obligation>

      <xsl:for-each select="Assertions/*">
	<xsl:call-template name="implantation_WDAssertions">
	  <xsl:with-param name="pos" select="position()"/> 
	</xsl:call-template>
      </xsl:for-each>

     <Proof_Obligation>
	<Comment></Comment>
        <Tag>WellDefinednessInitialisation</Tag>
        <Definition name="B definitions"/>
        <Definition name="ctx"/>
        <Definition name="mchcst"/>
        <Definition name="aprp"/>
        <Definition name="imlprp"/>
        <Definition name="imprp"/>
        <Definition name="imext"/>
        <Definition name="seext"/>
	<Goal goalTag="Initialisation is well defined">
          <Well_Definedness_Subst>
            <xsl:copy-of select="Initialisation/*"/>
          </Well_Definedness_Subst>
        </Goal>
      </Proof_Obligation>

      <xsl:for-each select="Local_Operations/Operation">
        <Proof_Obligation>
          <Comment>Operation locale dans une implementation</Comment>
          <Tag>WellDefinedness_Precondition_<xsl:value-of select="@name"/></Tag>
          <Definition name="B definitions"/>
          <Definition name="ctx"/>
          <Definition name="mchcst"/>
          <Definition name="aprp"/>
          <Definition name="imlprp"/>
          <Definition name="imprp"/>
          <Definition name="seext"/>
          <Hypothesis>
            <Nary_Pred>
              <xsl:attribute name="op">
                <xsl:value-of select="$and" />
              </xsl:attribute>
              <xsl:copy-of select="/Machine/Imports/Machine/Unrenamed_Invariant/*"/>
            </Nary_Pred>
          </Hypothesis>
	  <Goal goalTag="Operation precondition is well defined">
            <Well_Definedness_Pred>
              <xsl:copy-of select="Precondition/*"/>
            </Well_Definedness_Pred>
          </Goal>
        </Proof_Obligation>

        <Proof_Obligation>
          <Comment>Operation locale dans une implementation</Comment>
          <Tag>WellDefinedness_<xsl:value-of select="@name"/></Tag>
          <Definition name="B definitions"/>
          <Definition name="ctx"/>
          <Definition name="mchcst"/>
          <Definition name="aprp"/>
          <Definition name="imlprp"/>
          <Definition name="imprp"/>
          <Definition name="seext"/>
          <Hypothesis>
             <Nary_Pred>
               <xsl:attribute name="op">
                 <xsl:value-of select="$and"/>
              </xsl:attribute>
              <xsl:copy-of select="/Machine/Imports/Machine/Unrenamed_Invariant/*"/>
            </Nary_Pred>
          </Hypothesis>
          <Hypothesis><xsl:copy-of select="Precondition/*"/></Hypothesis>
	  <Goal goalTag="Operation is well defined">
            <Well_Definedness_Subst>
              <xsl:copy-of select="Body/*"/>
            </Well_Definedness_Subst>
          </Goal>
        </Proof_Obligation>
      </xsl:for-each>

      <xsl:for-each select="Operations/*">
          <Proof_Obligation>
          <Comment>Operations dans une implementation</Comment>
          <Tag>WellDefinedness_<xsl:value-of select="@name"/></Tag>
          <Definition name="B definitions"/>
          <Definition name="ctx"/>
          <Definition name="mchcst"/>
          <Definition name="aprp"/>
          <Definition name="imlprp"/>
          <Definition name="imprp"/>
          <Definition name="imext"/>
          <Definition name="seext"/>
          <Definition name="abs"/>
          <Definition name="inv"/>
          <Definition name="ass"/>
          <Hypothesis>
              <Nary_Pred>
                <xsl:attribute name="op">
                  <xsl:value-of select="$and"/>
                </xsl:attribute>
                <xsl:copy-of select="/Machine/Abstraction/Machine/Operations/Operation[@name=current()/@name]/Precondition/*"/>
                <xsl:copy-of select="/Machine/Abstraction/Machine/Operations/ImplicitOperation[@name=current()/@name]/Precondition/*"/>
                <xsl:copy-of select="Precondition/*"/>
              </Nary_Pred>
          </Hypothesis>
	  <Goal goalTag="Operation is well defined">
            <Well_Definedness_Subst>
              <xsl:copy-of select="Body/*"/>
            </Well_Definedness_Subst>
          </Goal>
        </Proof_Obligation>
      </xsl:for-each>

      <TypeInfos>
        <xsl:copy-of select="TypeInfos/*"/>
      </TypeInfos>

    </Proof_Obligations>
  </xsl:template>

  <xsl:template name="machine_Assertions">
    <xsl:param name="pos"/>
    <Proof_Obligation>
      <Comment>Assertion dans une machine abstraite</Comment>
      <Tag>AssertionLemmas</Tag>
      <Definition name="B definitions"/>
      <Definition name="ctx"/>
      <Definition name="cst"/>
      <Definition name="lprp"/>
      <Definition name="inprp"/>
      <Definition name="inext"/>
      <Definition name="seext"/>
      <Definition name="inv"/>
      <xsl:for-each select="../*">
        <xsl:if test="position() &lt; $pos">
          <Hypothesis>
            <xsl:copy-of select="."/>
          </Hypothesis>
        </xsl:if>
      </xsl:for-each>
      <Goal>
	<xsl:attribute name="goalTag">Assertion is verified</xsl:attribute>
	<xsl:copy-of select="."/>
      </Goal>
    </Proof_Obligation>

  </xsl:template>

  <xsl:template name="machine_WDAssertions">
    <xsl:param name="pos" />

    <Proof_Obligation>
      <Comment></Comment>
      <Tag>WellDefinednessAssertions</Tag>
      <Definition name="B definitions"/>
      <Definition name="ctx"/>
      <Definition name="cst"/>
      <Definition name="lprp"/>
      <Definition name="inprp"/>
      <Definition name="inext"/>
      <Definition name="seext"/>
      <Definition name="inv"/>
      <xsl:for-each select="../*">
        <xsl:if test="position() &lt; $pos">
          <Hypothesis>
            <xsl:copy-of select="."/>
          </Hypothesis>
        </xsl:if>
      </xsl:for-each>
      <Goal goalTag="Assertion is well defined">
        <Well_Definedness_Pred>
          <xsl:copy-of select="."/>
        </Well_Definedness_Pred>
      </Goal>
    </Proof_Obligation>
  </xsl:template>

  <xsl:template name="refinement_Assertions">
    <xsl:param name="pos"/>
    <Proof_Obligation>
      <Comment>Assertion dans un raffinement</Comment>
      <Tag>AssertionLemmas</Tag>
      <Definition name="B definitions"/>
      <Definition name="ctx"/>
      <Definition name="mchcst"/>
      <Definition name="aprp"/>
      <Definition name="lprp"/>
      <Definition name="inprp"/>
      <Definition name="inext"/>
      <Definition name="seext"/>
      <Definition name="abs"/>
      <Definition name="inv"/>
      <xsl:for-each select="../*">
        <xsl:if test="position() &lt; $pos">
          <Hypothesis>
            <xsl:copy-of select="."/>
          </Hypothesis>
        </xsl:if>
      </xsl:for-each>
      <Goal>
	<xsl:attribute name="goalTag">Assertion is verified</xsl:attribute>
	<xsl:copy-of select="."/>
      </Goal>
    </Proof_Obligation>
  </xsl:template>

  <xsl:template name="refinement_WDAssertions">
    <xsl:param name="pos"/>
    <Proof_Obligation>
      <Comment></Comment>
      <Tag>WellDefinednessAssertions</Tag>
      <Definition name="B definitions"/>
      <Definition name="ctx"/>
      <Definition name="mchcst"/>
      <Definition name="aprp"/>
      <Definition name="lprp"/>
      <Definition name="inprp"/>
      <Definition name="inext"/>
      <Definition name="seext"/>
      <Definition name="abs"/>
      <Definition name="inv"/>
      <xsl:for-each select="../*">
        <xsl:if test="position() &lt; $pos">
          <Hypothesis>
            <xsl:copy-of select="."/>
          </Hypothesis>
        </xsl:if>
      </xsl:for-each>
      <Goal goalTag="Assertion is well defined">
        <Well_Definedness_Pred>
          <xsl:copy-of select="."/>
        </Well_Definedness_Pred>
      </Goal>
    </Proof_Obligation>
  </xsl:template>

  <xsl:template name="implantation_Assertions">
    <xsl:param name="pos"/>
    <Proof_Obligation>
      <Comment>Assertion dans une implantation</Comment>
      <Tag>AssertionLemmas</Tag>
      <Definition name="B definitions"/>
      <Definition name="ctx"/>
      <Definition name="mchcst"/>
      <Definition name="aprp"/>
      <Definition name="imlprp"/>
      <Definition name="imprp"/>
      <Definition name="imext"/>
      <Definition name="seext"/>
      <Definition name="abs"/>
      <Definition name="inv"/>
      <xsl:for-each select="../*">
         <xsl:if test="position() &lt; $pos">
          <Hypothesis>
            <xsl:copy-of select="."/>
          </Hypothesis>
        </xsl:if>
      </xsl:for-each>
      <Goal>
	<xsl:attribute name="goalTag">Assertion is verified</xsl:attribute>
	<xsl:copy-of select="."/>
      </Goal>
    </Proof_Obligation>
  </xsl:template>

  <xsl:template name="implantation_WDAssertions">
    <xsl:param name="pos"/>
    <Proof_Obligation>
      <Comment></Comment>
      <Tag>WellDefinednessAssertions</Tag>
      <Definition name="B definitions"/>
      <Definition name="ctx"/>
      <Definition name="mchcst"/>
      <Definition name="aprp"/>
      <Definition name="imlprp"/>
      <Definition name="imprp"/>
      <Definition name="imext"/>
      <Definition name="seext"/>
      <Definition name="abs"/>
      <Definition name="inv"/>
      <xsl:for-each select="../*">
        <xsl:if test="position() &lt; $pos">
          <Hypothesis>
            <xsl:copy-of select="."/>
          </Hypothesis>
        </xsl:if>
      </xsl:for-each>
      <Goal goalTag="Assertion is well defined">
        <Well_Definedness_Pred>
          <xsl:copy-of select="."/>
        </Well_Definedness_Pred>
      </Goal>
    </Proof_Obligation>
  </xsl:template>

  <xsl:template name="refinement_merge">
    <xsl:param name="opName"/>
    <xsl:param name="refined_events"/>

    <Proof_Obligation>
      <Comment>Operations dans un raffinement. Cas d'une fusion</Comment>
      <Tag><xsl:value-of select="$opName"/></Tag>
      <Definition name="B definitions"/>
      <Definition name="ctx"/>
      <Definition name="mchcst"/>
      <Definition name="aprp"/>
      <Definition name="lprp"/>
      <Definition name="inprp"/>
      <Definition name="inext"/>
      <Definition name="seext"/>
      <Definition name="abs"/>
      <Definition name="inv"/>
      <Definition name="ass"/>
      <Goal>
	<xsl:attribute name="goalTag">Refinement is correct</xsl:attribute>
	<Sub_Calculus>
	  <xsl:copy-of select="/Machine/Operations/Operation[@name=$opName]/Body/*"/>
	  <xsl:copy-of select="/Machine/Operations/ImplicitOperation[@name=$opName]/Body/*"/>
	  <Binary_Pred op="=>">
	    <Nary_Pred>
	      <xsl:attribute name="op">
		<xsl:value-of select="$and" />
	      </xsl:attribute>
	      <xsl:copy-of select="Includes/Machine/Invariant/*"/>
	      <xsl:copy-of select="Includes/Machine/Assertions/*"/>
	    </Nary_Pred>
	    <Not>
	      <Sub_Calculus>
		<Nary_Sub op='CHOICE'>
		  <xsl:for-each select="$refined_events">
		    <xsl:call-template name="copy_abstract_body">
		      <xsl:with-param name="abstractOpName" select="./@value"/>
		    </xsl:call-template>
		  </xsl:for-each>
		</Nary_Sub>
		<Not>
		  <Nary_Pred>
		    <xsl:attribute name="op">
		      <xsl:value-of select="$and" />
		    </xsl:attribute>
		    <xsl:copy-of select="/Machine/Invariant/*"/>
		  </Nary_Pred>
		</Not>
	      </Sub_Calculus>
	    </Not>
	  </Binary_Pred>
	</Sub_Calculus>
      </Goal>
    </Proof_Obligation>
  </xsl:template>

  <xsl:template name="refinement_outputParamaters">
    <xsl:param name="opName"/>
    <xsl:param name="pos"/>
    <Exp_Comparison  op="=">
      <xsl:copy-of select="/Machine/Abstraction/Machine[1]/Operations/Operation[@name=$opName]/Output_Parameters/Id[position()=$pos]"/>
      <xsl:copy-of select="/Machine/Abstraction/Machine[1]/Operations/ImplicitOperation[@name=$opName]/Output_Parameters/Id[position()=$pos]"/>
      <xsl:element name="Id">
        <xsl:attribute name="value"><xsl:copy-of select="@value"/></xsl:attribute>
        <xsl:attribute name="suffix"><xsl:copy-of select="@suffix"/></xsl:attribute>
        <xsl:attribute name="typref"><xsl:copy-of select="@typref"/></xsl:attribute>
        <xsl:copy-of select="*"/>
      </xsl:element>
    </Exp_Comparison>
  </xsl:template>

  <xsl:template name="implantation_outputParamaters">
    <xsl:param name="opName"/>
    <xsl:param name="pos"/>
    <Exp_Comparison  op="=">
      <xsl:copy-of select="/Machine/Operations/Operation[@name=$opName]/Output_Parameters/Id[position()=$pos]"/>
      <xsl:copy-of select="/Machine/Operations/ImplicitOperation[@name=$opName]/Output_Parameters/Id[position()=$pos]"/>
      <xsl:element name="Id">
        <xsl:attribute name="value"><xsl:copy-of select="@value"/></xsl:attribute>
        <xsl:attribute name="suffix"><xsl:copy-of select="@suffix"/></xsl:attribute>
        <xsl:attribute name="typref"><xsl:copy-of select="@typref"/></xsl:attribute>
        <xsl:copy-of select="*"/>
      </xsl:element>
    </Exp_Comparison>
  </xsl:template>

  <xsl:template match="Operation|ImplicitOperation" name="Operation_refinement">
    <xsl:param name="opName"/>
    <xsl:param name="abstractOpName" select="$opName"/>
    <Proof_Obligation>
      <Comment>Operations dans un raffinement / Precondition</Comment>
      <Tag><xsl:value-of select="$opName"/></Tag>
      <Definition name="B definitions"/>
      <Definition name="ctx"/>
      <Definition name="mchcst"/>
      <Definition name="aprp"/>
      <Definition name="lprp"/>
      <Definition name="inprp"/>
      <Definition name="inext"/>
      <Definition name="seext"/>
      <Definition name="abs"/>
      <Definition name="inv"/>
      <Definition name="ass"/>
      <Hypothesis>
        <Nary_Pred>
          <xsl:attribute name="op">
            <xsl:value-of select="$and" />
          </xsl:attribute>
	  <xsl:copy-of select="/Machine/Abstraction/Machine/Operations/Operation[@name=$abstractOpName]/Precondition/*"/>
	  <xsl:copy-of select="/Machine/Abstraction/Machine/Operations/ImplicitOperation[@name=$abstractOpName]/Precondition/*"/>
	</Nary_Pred>
      </Hypothesis>
      <Goal>
	<xsl:attribute name="goalTag">Precondition can be deduced</xsl:attribute>
	<Binary_Pred op="=>">
	  <Nary_Pred>
	    <xsl:attribute name="op">
	      <xsl:value-of select="$and"/>
	    </xsl:attribute>
            <xsl:for-each select="/Machine/Operations/Operation[@name=$opName]/Output_Parameters/*">
              <xsl:call-template name="refinement_outputParamaters">
                <xsl:with-param name="opName" select="$abstractOpName"/>
                <xsl:with-param name="pos" select="position()"/>
              </xsl:call-template>
            </xsl:for-each>
	  </Nary_Pred>
	  <Nary_Pred>
	    <xsl:attribute name="op">
	      <xsl:value-of select="$and" />
	    </xsl:attribute>
	    <xsl:copy-of select="/Machine/Operations/Operation[@name=$opName]/Precondition/*"/>
	  </Nary_Pred>
	</Binary_Pred>
      </Goal>
    </Proof_Obligation>

    <Proof_Obligation>
      <Comment>Operations dans un raffinement</Comment>
      <Tag><xsl:value-of select="$opName"/></Tag>
      <Definition name="B definitions"/>
      <Definition name="ctx"/>
      <Definition name="mchcst"/>
      <Definition name="aprp"/>
      <Definition name="lprp"/>
      <Definition name="inprp"/>
      <Definition name="inext"/>
      <Definition name="seext"/>
      <Definition name="abs"/>
      <Definition name="inv"/>
      <Definition name="ass"/>
      <Hypothesis>
        <Nary_Pred>
	  <xsl:attribute name="op">
	    <xsl:value-of select="$and" />
	  </xsl:attribute>
	  <xsl:copy-of select="/Machine/Abstraction/Machine/Operations/Operation[@name=$abstractOpName]/Precondition/*"/>
	  <xsl:copy-of select="/Machine/Abstraction/Machine/Operations/ImplicitOperation[@name=$abstractOpName]/Precondition/*"/>
	</Nary_Pred>
      </Hypothesis>
      <Goal>
	<xsl:attribute name="goalTag">Refinement is correct</xsl:attribute>
	<Binary_Pred op="=>">
	  <Nary_Pred>
	    <xsl:attribute name="op">
	      <xsl:value-of select="$and" />
	    </xsl:attribute>
	    <xsl:copy-of select="/Machine/Abstraction/Machine/Operations/Operation[@name=$abstractOpName]/Precondition/*"/>
	    <xsl:copy-of select="/Machine/Abstraction/Machine/Operations/ImplicitOperation[@name=$abstractOpName]/Precondition/*"/>
	    <xsl:copy-of select="/Machine/Operations/Operation[@name=$opName]/Precondition/*"/>
            <xsl:for-each select="/Machine/Operations/Operation[@name=$opName]/Output_Parameters/*" >
              <xsl:call-template name="refinement_outputParamaters">
                <xsl:with-param name="opName" select="$abstractOpName"/>
                <xsl:with-param name="pos" select="position()"/>
              </xsl:call-template>
            </xsl:for-each>
	  </Nary_Pred>
	  <Sub_Calculus>
	    <xsl:copy-of select="/Machine/Operations/Operation[@name=$opName]/Body/*"/>
	    <xsl:copy-of select="/Machine/Operations/ImplicitOperation[@name=$opName]/Body/*"/>
	    <Binary_Pred op="=>">
	      <Nary_Pred>
		<xsl:attribute name="op">
		  <xsl:value-of select="$and" />
		</xsl:attribute>
		<xsl:copy-of select="Includes/Machine/Invariant/*"/>
		<xsl:copy-of select="Includes/Machine/Assertions/*"/>
	      </Nary_Pred>
	      <Not>
		<Sub_Calculus>
		  <xsl:call-template name="copy_abstract_body">
		    <xsl:with-param name="abstractOpName" select="$abstractOpName"/>
		  </xsl:call-template>
		  <Not>
		    <Nary_Pred>
		      <xsl:attribute name="op">
			<xsl:value-of select="$and" />
		      </xsl:attribute>
		      <xsl:copy-of select="/Machine/Invariant/*"/>
                      <xsl:for-each select="/Machine/Operations/Operation[@name=$opName]/Output_Parameters/*">
                        <xsl:call-template name="refinement_outputParamaters">
                          <xsl:with-param name="opName" select="$abstractOpName"/>
                          <xsl:with-param name="pos" select="position()"/>
                        </xsl:call-template>
                      </xsl:for-each>
		    </Nary_Pred>
		  </Not>
		</Sub_Calculus>
	      </Not>
	    </Binary_Pred>
	  </Sub_Calculus>
	</Binary_Pred>
      </Goal>
    </Proof_Obligation>
  </xsl:template>

  <xsl:template match="Operation|ImplicitOperation" name="Operation_implantation">
    <xsl:param name="opName"/>
    <xsl:param name="abstractOpName" select="$opName"/>
	<!-- cette PO est actuellement inutile, les préconditions etant interdites dans les implementations.
	     Elle devra etre reactivée si l'on décide de supprimer cette interdiction.
    <Proof_Obligation>
      <Comment>Operations dans une implantation / Precondition</Comment>
      <Tag><xsl:value-of select="$opName"/></Tag>
      <Definition name="B definitions"/>
      <Definition name="ctx"/>
      <Definition name="mchcst"/>
      <Definition name="aprp"/>
      <Definition name="imlprp"/>
      <Definition name="imprp"/>
      <Definition name="imext"/>
      <Definition name="seext"/>
      <Definition name="abs"/>
      <Definition name="inv"/>
      <Definition name="ass"/>
      <Hypothesis>
        <Nary_Pred>
          <xsl:attribute name="op">
            <xsl:value-of select="$and" />
          </xsl:attribute>
          <xsl:copy-of select="/Machine/Abstraction/Machine/Operations/Operation[@name=$abstractOpName]/Precondition/*"/>
          <xsl:copy-of select="/Machine/Abstraction/Machine/Operations/ImplicitOperation[@name=$abstractOpName]/Precondition/*"/>
        </Nary_Pred>
      </Hypothesis>
      <Goal goalTag="Precondition is verified">
        <Binary_Pred op="=>">
          <Nary_Pred>
            <xsl:attribute name="op">
                 <xsl:value-of select="$and" />
              </xsl:attribute>
              <xsl:copy-of select="/Machine/Abstraction/Machine/Operations/Operation[@name=$opName]/Precondition/*"/>
            <xsl:copy-of select="/Machine/Abstraction/Machine/Operations/ImplicitOperation[@name=$abstractOpName]/Precondition/*"/>
            <xsl:apply-templates select="Output_Parameters/*" mode="implantation_outputParamaters">
              <xsl:with-param name="opName" select="$opName"/>
              <xsl:with-param name="pos" select="position()"/>
            </xsl:apply-templates>
          </Nary_Pred>
	  <Nary_Pred>
	    <xsl:attribute name="op">
	      <xsl:value-of select="$and" />
	    </xsl:attribute>
	    <xsl:copy-of select="/Machine/Operations/Operation[@name=$opName]/Precondition/*"/>
	  </Nary_Pred>
	</Binary_Pred>
      </Goal>
    </Proof_Obligation>
    -->

    <Proof_Obligation>
      <Comment>Operations dans une implantation</Comment>
      <Tag><xsl:value-of select="$opName"/></Tag>
      <Definition name="B definitions"/>
      <Definition name="ctx"/>
      <Definition name="mchcst"/>
      <Definition name="aprp"/>
      <Definition name="imlprp"/>
      <Definition name="imprp"/>
      <Definition name="imext"/>
      <Definition name="seext"/>
      <Definition name="abs"/>
      <Definition name="inv"/>
      <Definition name="ass"/>
      <Goal>
	<xsl:attribute name="goalTag">Refinement is correct</xsl:attribute>
	<Sub_Calculus>
	  <Nary_Sub op = ";">
	    <xsl:copy-of select="/Machine/Renaming_Implicit_Gluing/*"/>
	  </Nary_Sub>
	  <Binary_Pred op="=>">
	    <Nary_Pred>
	      <xsl:attribute name="op">
		<xsl:value-of select="$and" />
	      </xsl:attribute>
	      <xsl:copy-of select="/Machine/Abstraction/Machine/Operations/Operation[@name=$abstractOpName]/Precondition/*"/>
	      <xsl:copy-of select="/Machine/Abstraction/Machine/Operations/ImplicitOperation[@name=$abstractOpName]/Precondition/*"/>
	      <xsl:copy-of select="/Machine/Operations/Operation[@name=$opName]/Precondition/*"/>
	      <xsl:copy-of select="/Machine/Operations/ImplicitOperation[@name=$opName]/Precondition/*"/>
	      <xsl:copy-of select="/Machine/Local_Operations/Operation[@name=$opName]/Precondition/*"/>
	      <xsl:for-each select="/Machine/Operations[@name=$opName]/Output_Parameters/*">
                <xsl:call-template name="implantation_outputParamaters">
                  <xsl:with-param name="opName" select="$opName"/>
                  <xsl:with-param name="pos" select="position()"/>
                </xsl:call-template>
              </xsl:for-each>
	    </Nary_Pred>
	    <Sub_Calculus>
	      <xsl:for-each select="/Machine/Operations/Operation[@name=$opName]/Body/*">
		<xsl:attribute name="overflow">
		  <xsl:value-of select="'true'" />
		</xsl:attribute>
	      </xsl:for-each>
	      <xsl:for-each select="/Machine/Operations/ImplicitOperation[@name=$opName]/Body/*">
		<xsl:attribute name="overflow">
		  <xsl:value-of select="'false'" />
		</xsl:attribute>
	      </xsl:for-each>
	      <Nary_Sub op = ";">
		<xsl:copy-of select="/Machine/Operations/Operation[@name=$opName]/Body/*"/>
		<xsl:copy-of select="/Machine/Operations/ImplicitOperation[@name=$opName]/Body/*"/>
	      </Nary_Sub>
	      <Binary_Pred op="=>">
		<Nary_Pred>
		  <xsl:attribute name="op">
		    <xsl:value-of select="$and" />
		  </xsl:attribute>
		  <xsl:copy-of select="Includes/Machine/Invariant/*"/>
		  <xsl:copy-of select="Includes/Machine/Assertions/*"/>
		</Nary_Pred>
		<Not>
		  <Sub_Calculus>
		    <xsl:copy-of select="/Machine/Abstraction/Machine[1]/Operations/Operation[@name=$abstractOpName]/Body/*"/>
		    <Not>
		      <Nary_Pred>
			<xsl:attribute name="op">
			  <xsl:value-of select="$and" />
			</xsl:attribute>
			<xsl:copy-of select="/Machine/Invariant/*"/>
                        <xsl:for-each select="/Machine/Operations/Operation[@name=$opName]/Output_Parameters/*">
                          <xsl:call-template name="implantation_outputParamaters">
                            <xsl:with-param name="opName" select="$opName"/>
                            <xsl:with-param name="pos" select="position()"/>
                          </xsl:call-template>
                        </xsl:for-each>
		      </Nary_Pred>
		    </Not>
		  </Sub_Calculus>
		</Not>
	      </Binary_Pred>
	    </Sub_Calculus>
	  </Binary_Pred>
	</Sub_Calculus>
      </Goal>
    </Proof_Obligation>
  </xsl:template>

  <xsl:template match="Operation|ImplicitOperation" name="local_implantation">
    <xsl:param name="opName"/>
	<!-- cette PO est actuellement inutile, les préconditions etant interdites dans les implementations.
	     Elle devra etre reactivée si l'on décide de supprimer cette interdiction.
    <Proof_Obligation>
      <Comment>Operations dans une implantation / Precondition</Comment>
      <Tag>refinement_of_<xsl:value-of select="$opName"/></Tag>
      <Definition name="B definitions"/>
      <Definition name="ctx"/>
      <Definition name="mchcst"/>
      <Definition name="aprp"/>
      <Definition name="imlprp"/>
      <Definition name="imprp"/>
      <Definition name="imext"/>
      <Definition name="seext"/>
      <Definition name="abs"/>
      <Definition name="inv"/>
      <Definition name="ass"/>
      <Goal goalTag="Precondition is verified">
        <Binary_Pred op="=>">
          <Nary_Pred>
            <xsl:attribute name="op">
              <xsl:value-of select="$and" />
              </xsl:attribute>
            <xsl:apply-templates select="Output_Parameters/*" mode="implantation_outputParamaters">
              <xsl:with-param name="opName" select="$opName"/>
              <xsl:with-param name="pos" select="position()"/>
            </xsl:apply-templates>
          </Nary_Pred>
          <Nary_Pred>
            <xsl:attribute name="op">
              <xsl:value-of select="$and" />
            </xsl:attribute>
            <xsl:copy-of select="/Machine/Operations/Operation[@name=$opName]/Precondition/*"/>
          </Nary_Pred>
        </Binary_Pred>
      </Goal>
    </Proof_Obligation>
    -->

    <Proof_Obligation>
      <Comment>Operation locale dans une implantation</Comment>
      <Tag>refinement_of_<xsl:value-of select="$opName"/></Tag>
      <Definition name="B definitions"/>
      <Definition name="ctx"/>
      <Definition name="mchcst"/>
      <Definition name="aprp"/>
      <Definition name="imlprp"/>
      <Definition name="imprp"/>
      <Definition name="imext"/>
      <Definition name="seext"/>
      <Definition name="abs"/>
      <Definition name="ass"/>
      <Hypothesis>
	<Nary_Pred>
	  <xsl:attribute name="op">
	    <xsl:value-of select="$and" />
	  </xsl:attribute>
	  <xsl:copy-of select="/Machine/Local_Invariant/*"/>
	</Nary_Pred>
      </Hypothesis>
      <Goal>
	<xsl:attribute name="goalTag">Refinement is correct</xsl:attribute>
	<Sub_Calculus>
	  <xsl:attribute name="overflow">
	    <xsl:value-of select="'true'" />
	  </xsl:attribute>
	  <Nary_Sub op = ";">
	    <xsl:copy-of select="/Machine/Renaming_Implicit_Gluing/*"/>
	  </Nary_Sub>
	  <Binary_Pred op="=>">
	    <Nary_Pred>
	      <xsl:attribute name="op">
		<xsl:value-of select="$and" />
	      </xsl:attribute>
	      <xsl:copy-of select="/Machine/Local_Operations/Operation[@name=$opName]/Precondition/*"/>
	      <xsl:for-each select="Output_Parameters/*">
                <xsl:call-template name="implantation_outputParamaters">
                  <xsl:with-param name="opName" select="$opName"/>
                  <xsl:with-param name="pos" select="position()"/>
                </xsl:call-template>
              </xsl:for-each>
	    </Nary_Pred>
	    <Sub_Calculus>
	      <xsl:attribute name="overflow">
		<xsl:value-of select="'true'" />
	      </xsl:attribute>
	      <Nary_Sub op = ";">
		<xsl:copy-of select="/Machine/Operations/Operation[@name=$opName]/Body/*"/>
	      </Nary_Sub>
	      <Binary_Pred op="=>">
		<Nary_Pred>
		  <xsl:attribute name="op">
		    <xsl:value-of select="$and" />
		  </xsl:attribute>
		  <xsl:copy-of select="Includes/Machine/Assertions/*"/>
		</Nary_Pred>
		<Not>
		  <Sub_Calculus>
		    <xsl:copy-of select="Body/*"/>
		    <Not>
		      <Nary_Pred>
			<xsl:attribute name="op">
			  <xsl:value-of select="$and" />
			</xsl:attribute>
			<xsl:copy-of select="/Machine/Local_Invariant/*"/>
			<xsl:for-each select="Output_Parameters/*">
                          <xsl:call-template name="implantation_outputParamaters">
                            <xsl:with-param name="opName" select="$opName"/>
                            <xsl:with-param name="pos" select="position()"/>
                          </xsl:call-template>
                        </xsl:for-each>
		      </Nary_Pred>
		    </Not>
		  </Sub_Calculus>
		</Not>
	      </Binary_Pred>
	    </Sub_Calculus>
	  </Binary_Pred>
	</Sub_Calculus>
      </Goal>
    </Proof_Obligation>
  </xsl:template>

  <xsl:template name="copy_abstract_body">
    <xsl:param name="abstractOpName"/>

    <xsl:choose>
      <xsl:when test = "boolean(
                        /Machine/Abstraction/Machine[1]/Operations/Operation[@name=$abstractOpName] |
                        /Machine/Abstraction/Machine[1]/Operations/ImplicitOperation[@name=$abstractOpName])"
                >
        <xsl:copy-of select="/Machine/Abstraction/Machine[1]/Operations/Operation[@name=$abstractOpName]/Body/*"/>
        <xsl:copy-of select="/Machine/Abstraction/Machine[1]/Operations/ImplicitOperation[@name=$abstractOpName]/Body/*"/>
      </xsl:when>
      <xsl:otherwise>
        <Skip/>
      </xsl:otherwise>
    </xsl:choose>

  </xsl:template>

</xsl:stylesheet>
